import os, sysconfig
import distutils.sysconfig as distutils_sysconfig
from .mypyc_aot import set_default_compiler

__all__ = ["init_custom_unix_compiler", "restore_default_compiler"]
_original_env = {}

def init_custom_unix_compiler(cc: str, cxx: str):
    # 自定义c/c++编译器，不依赖sysconfig的配置
    # cc为"gcc" / "clang"，cxx为"g++" / "clang++"
    env_vars = {
        "CC": cc,
        "CXX": cxx,
        "LDSHARED": f"{cc} -shared",
        "AR": "ar",
        "ARFLAGS": ""
    }
    sysconfig_vars = {
        "CCSHARED": "-shared",
        "CFLAGS": "-s -O2 -Wall -pipe"
    }

    for var in env_vars.keys():
        if var in os.environ:
            _original_env.setdefault(var, os.environ[var])
        else:
            _original_env.setdefault(var, None)
    os.environ.update(env_vars)

    if distutils_sysconfig._config_vars is None:
        distutils_sysconfig._config_vars = sysconfig.get_config_vars().copy()
    distutils_sysconfig._config_vars.update(sysconfig_vars)
    set_default_compiler("unix")

def restore_default_compiler():
    for var, value in _original_env.items():
        if value is None:
            if var in os.environ:
                del os.environ[var]
        else:
            os.environ[var] = value
    _original_env.clear()
    
    distutils_sysconfig._config_vars = sysconfig.get_config_vars().copy()
    set_default_compiler(None)



from .mypyc_aot import *
from .unix_compiler_util import *
from .compiler import *

__all__ = [name for name in globals() if not name.startswith("_")]

try:
    from .ipy_integration import load_ipython_extension
except ImportError: # 未安装IPython
    pass

__version__ = "1.0.3"


import sys, time
from mypy_extensions import i64
from mypyc_aot import Compiler

compiler = Compiler(globals())
if "--gcc" in sys.argv[1:]:
    compiler.use_custom_compiler("gcc", "g++")
    sys.argv.remove("--gcc")

@compiler.aot
def compute(x: i64) -> i64:
    sum: i64 = 0
    for i in range(x):
        sum += i
    return sum

def compute_no_mypyc(x: i64) -> i64:
    sum: i64 = 0
    for i in range(x):
        sum += i
    return sum

compiler.start_compilation_thread()

def main():
    x = int(sys.argv[1]) if len(sys.argv) >= 2 else 50000000

    start = time.perf_counter()
    result_normal = compute_no_mypyc(x)
    normal_time = time.perf_counter() - start

    compute(0) # 等待编译完成
    start = time.perf_counter()
    result_mypyc = compute(x)
    mypyc_time = time.perf_counter() - start
    
    speedup = normal_time / mypyc_time
    print(f"Result: {result_mypyc}")
    print(f"mypyc: {mypyc_time:.3f}s")
    print(f"Pure Python: {normal_time:.3f}s")
    print(f"Speedup: {speedup:.2f}x")
    print(f"Results match: {result_mypyc == result_normal}")

if __name__ == "__main__":main()


import re, ast, inspect, warnings, functools, itertools, threading
from typing import Any, _Final
from .mypyc_aot import mypyc_aot_nocache, mypyc_aot, CACHE_DIR
from .unix_compiler_util import init_custom_unix_compiler, restore_default_compiler

__all__ = ["Compiler"]

AOT_PATTERN = re.compile(r"^@([0-9A-Za-z_]+?)\.aot[^\n]*?\n")
IGNORED_NAMES = ["__annotations__", "__builtins__", "__cached__",
                 "__loader__", "__spec__"]
BASIC_TYPES = {int, float, str, bytes, bytearray, list,
               tuple, dict, set, type(None), type(range(0)), slice}
REPR_TYPES = {re.Pattern}

def is_representable(obj): # 检查对象能否被repr()表示
    # obj为非闭包类
    if isinstance(obj, (type, _Final)) and \
        obj.__name__ == obj.__qualname__: # type: ignore[union-attr]
        return True
    elif type(obj) in BASIC_TYPES: # 不使用isinstance（由于不能是基本类型子类）
        if type(obj) in (list, tuple, dict, set, frozenset):
            if isinstance(obj, dict):
                obj = itertools.chain(obj.keys(), obj.values())
            return all(is_representable(sub) for sub in obj)
        return True
    elif type(obj) in REPR_TYPES:
        return True
    return False

class ReprWrapper:
    def __init__(self, obj, repr_func):
        self.obj = obj
        self.repr_func = repr_func
    def __repr__(self) -> str:
        return self.repr_func(self.obj)

def get_source(function_or_type) -> str | None:
    try:
        source = inspect.getsource(function_or_type)
    except OSError:
        return None
    if source.startswith(" "):
        raise ValueError(f"""Cannot compile methods or closure functions \
({function_or_type.__qualname__}). Use @aot for outermost codes.""") # type: ignore[return]
    source = re.sub(AOT_PATTERN, "", source) # 去除compiler.aot
    return source

def parse_name_from_source(source: str) -> str:
    parsed = ast.parse(source)
    if len(parsed.body) != 1 or not \
        isinstance(parsed.body[0], (ast.FunctionDef, ast.ClassDef)):
        raise ValueError("not a function or class")
    return parsed.body[0].name

class Compiler:
    _scope: dict
    _added_symbols: set[str]
    _cache_dir: str | None
    _quiet: bool
    _compiler: str | None
    _ignore_import_not_found: bool
    _ignore_self: bool
    _custom_compiler: dict[str, str] | None
    _codes: list[str]
    _comp_thread: threading.Thread | None
    _modules: set[str] # add_symbols_from中临时使用
    _exported_names: set[str]
    _orig_module_names: dict[str, str] # 函数/类编译前的模块名（__module__）
    _has_class_in_source: bool
    def __init__(self, scope: dict, ignored_vars: list[str] | None = None,
                 cache_dir=CACHE_DIR, quiet=True, compiler=None,
                 no_symbol_warnings=False, ignore_import_not_found=True,
                 ignore_self=True, **kw):
        self._scope = scope
        self._cache_dir = cache_dir
        self._quiet = quiet
        self._compiler = compiler
        self._ignore_import_not_found = ignore_import_not_found
        self._ignore_self = ignore_self
        self._kw_options = kw
        self._custom_compiler = None
        self._codes = []
        self._comp_thread = None
        self._added_symbols = set()
        self._modules = set()
        self._exported_names = set()
        self._orig_module_names = {}
        self._has_class_in_source = False
        self.add_symbols_from(self._scope, ignored_vars, no_symbol_warnings)
    def add_symbols_from(self, scope: dict, ignored_vars: list[str] | None = None,
                         no_warnings=False):
        if ignored_vars is None: ignored_vars = []
        annotations: dict[str, Any] = self._scope.get("__annotations__",{})

        for name, value in list(scope.items()): # scope可能会改变大小
            if name in IGNORED_NAMES or name in ignored_vars:
                continue
            self._added_symbols.add(name)
            if name in ["__doc__", "__package__"] and value is None:
                continue # 避免Incompatible types in assignment

            if inspect.ismodule(value):
                # 忽略mypyc_aot模块自身
                if self._ignore_self and value.__name__.startswith("mypyc_aot"):
                    continue
                if value.__name__ == name:
                    self._modules.add(name)
                else:
                    comment = " # type: ignore[import-not-found]" \
                              if self._ignore_import_not_found else ""
                    self._codes.append(f"import {value.__name__} as {name}{comment}")
            elif isinstance(value, type) or inspect.isfunction(value) \
                or inspect.isbuiltin(value):
                if self._ignore_self and value.__module__.startswith("mypyc_aot"):
                    continue
                if value.__module__ == scope["__name__"]:
                    source = get_source(value)
                    if source is not None:
                        self._codes.append(source)
                        continue
                self._orig_module_names[name] = value.__module__
                comment = " # type: ignore[import-not-found]" \
                          if self._ignore_import_not_found else ""
                self._codes.append(
                    f"from {value.__module__} import {value.__name__} as {name}{comment}")
            elif is_representable(value):
                type_ = None
                if name in annotations:
                    type_ = self._repr(annotations[name])
                elif isinstance(value, _Final): # typing模块中的类型
                    self._modules.add("typing")
                    type_ = "typing.Any"

                if type_ is not None:
                    self._codes.append(f"{name}: {type_} = {self._repr(value)}")
                else:
                    self._codes.append(f"{name} = {self._repr(value)}")
            elif not no_warnings:
                warnings.warn(f"skipping variable {name} ({repr(value)[:100]})")

        self._codes = [f"import {mod}" for mod in sorted(self._modules)] + self._codes
        self._modules.clear()
    def _repr(self, obj) -> str:
        if obj is type(None):
            return "type(None)"
        if isinstance(obj, (type, _Final)):
            self._modules.add(obj.__module__) # type: ignore[union-attr]
            return f"{obj.__module__}.{obj.__qualname__}" # type: ignore[union-attr]
        if type(obj) in (list, tuple, set, frozenset):
            return repr(type(obj)(ReprWrapper(item, self._repr) for item in obj))
        if type(obj) is dict:
            return repr({ReprWrapper(k, self._repr) :ReprWrapper(v, self._repr)
                         for k,v in obj.items()})
        return repr(obj)

    def aot(self, function_or_type):
        name = function_or_type.__name__
        source = get_source(function_or_type)
        if source is None:
            raise ValueError(f"{name} has no sources")
        if name not in self._added_symbols: # 避免重复定义
            self._codes.append(source)
        self._exported_names.add(name)
        self._orig_module_names[name] = self._scope["__name__"]

        if isinstance(function_or_type, type):
            self._has_class_in_source = True
        if self._has_class_in_source:
            return None # 调用compile()后再写回self._scope

        @functools.wraps(function_or_type) # type: ignore[arg-type]
        def wrapper(*args, **kw):
            if self._comp_thread is None:
                raise RuntimeError(f"Call start_compilation() before calling {name}")
            self._comp_thread.join()
            if self._scope[name] is wrap:
                raise ValueError("compilation is not successful")
            return self._scope[name](*args, **kw)

        wrap = wrapper # wrap: 避免attribute 'wrapper' of 'aot_Compiler_env' undefined
        return wrapper
    def add_func_or_class(self, source: str):
        name = parse_name_from_source(source)
        self._exported_names.add(name)
        self._orig_module_names[name] = self._scope["__name__"]
        self._codes.append(source)
    def add_source(self, source: str):
        self._codes.append(source)

    def get_source_code(self):
        return "\n".join(self._codes)
    def compile(self):
        try:
            if self._custom_compiler is not None:
                init_custom_unix_compiler(self._custom_compiler["cc"],
                                          self._custom_compiler["cxx"])
            if self._cache_dir is None:
                module = mypyc_aot_nocache(self.get_source_code(), quiet=self._quiet,
                                        compiler=self._compiler, **self._kw_options)
            else:
                module = mypyc_aot(self.get_source_code(), cache_dir=self._cache_dir,
                                quiet=self._quiet, compiler=self._compiler,
                                **self._kw_options)
            # 写回self._scope命名空间
            for name in self._exported_names:
                self._scope[name] = getattr(module, name)
                if name in self._orig_module_names:
                    try: # 尝试保持编译后模块名不变
                        self._scope[name].__module__ = self._orig_module_names[name]
                    except Exception: pass
        finally:
            if self._custom_compiler is not None:
                restore_default_compiler()
    def start_compilation_thread(self):
        if self._has_class_in_source:
            raise ValueError("""start_compilation_thread() is disabled when \
classes are used. Use compile() instead.""")
        self._comp_thread = threading.Thread(target=self.compile, daemon=True)
        self._comp_thread.start()
    def use_custom_compiler(self, cc: str, cxx: str):
        self._custom_compiler = {"cc": cc, "cxx": cxx}

# 自举
#if hasattr(get_source, "__code__"):
#    c = Compiler(globals(), ignore_self=False)
#    Compiler = c.aot(Compiler)
#    c.compile()



import re
from IPython.core.magic import Magics, magics_class, cell_magic
from .compiler import Compiler

IGNORED_NAMES = ["get_ipython", "exit", "quit", "In", "Out",
                 "_ih", "_oh", "_dh", "_", "__", "___", "_i",
                 "_ii", "_iii", "__session__", "__builtin__"]
IGNORED_VAR_PATTERN = re.compile(r"_\d*?|_i\d*?")

@magics_class
class MyPyCAotMagics(Magics):
    def __init__(self, shell):
        super().__init__(shell)
    def _get_ignored_names(self):
        ignored = IGNORED_NAMES.copy()
        for name in self.shell.user_ns:
            if re.match(IGNORED_VAR_PATTERN, name):
                ignored.append(name)
        return ignored
    @cell_magic
    def mypyc_aot(self, line, cell):
        compiler = Compiler(self.shell.user_ns,
                            self._get_ignored_names())
        compiler.add_func_or_class(cell)
        compiler.compile()

def load_ipython_extension(ipython):
    ipython.register_magics(MyPyCAotMagics)



import sys, os, io, string, base64, random
import pickle, sysconfig, tempfile
from hashlib import sha256
import subprocess
try:
    import distutils.compilers.C.base as _base
except ImportError:
    import distutils.ccompiler as _base
from setuptools import setup
from mypy.defaults import CACHE_DIR as _mypy_cache_basename
from mypyc.build import mypycify

__all__ = ["mypyc_aot_nocache", "mypyc_aot"]
_home = os.getenv("USERPROFILE") if sys.platform == "win32" \
        else os.getenv("HOME")
CACHE_DIR = os.path.join(_home, ".mypyc_aot_cache")
try:
    __import__("zstandard")
    DEFAULT_COMP_METHOD = "zstandard"
except ImportError:
    try:
        __import__("zlib")
        DEFAULT_COMP_METHOD = "zlib"
    except ImportError:
        DEFAULT_COMP_METHOD = None

_config = {}

class NoCompression:
    @staticmethod
    def compress(data: bytes) -> bytes: return data
    @staticmethod
    def decompress(data: bytes) -> bytes: return data

def rand_name(length):
    chars = string.ascii_letters + string.digits
    return "".join(chars[random.randrange(len(chars))] for _ in range(length))

def set_default_compiler(compiler: str | None):
    if compiler is None:
        del _config["default_compiler"]
    else: _config["default_compiler"] = compiler

def get_module_path(module_name):
    for finder in sys.meta_path:
        spec = finder.find_spec(module_name, sys.path)
        if spec is not None:
            return spec.origin
    return None

def captured_popen(stream):
    class Popen(subprocess.Popen):
        def __init__(self, *args, **kw):
            kw.setdefault("stdout", subprocess.PIPE)
            kw.setdefault("stderr", subprocess.PIPE)
            return super().__init__(*args, **kw)
        def wait(self, *args, **kw):
            result = super().wait(*args, **kw)
            if not self.stdout.closed and not self.stderr.closed:
                # _communicate: 避免递归
                stdout, stderr = self._communicate(None, None, None)
                if stdout is not None:
                    stream.write(stdout.decode("utf-8", "replace"))
                if stderr is not None:
                    stream.write(stderr.decode("utf-8", "replace"))
            return result
    return Popen

def mypyc_aot_nocache(pycode: str, prefix="mypyc_aot", cache_dir = None,
                      quiet=True, compiler: str | None = None,
                      opt_level="3", strict_dunder_typing=True,
                      experimental_features=False):
    if cache_dir is None:
        temp_dir = tempfile.gettempdir()
    else:
        temp_dir = os.path.join(cache_dir, "cache")
    py_path = os.path.join(temp_dir, f"{rand_name(16)}-mypyc-aot")
    os.makedirs(py_path, exist_ok=True)

    mod_name = f"{prefix}_{rand_name(16)}"
    pyfile = os.path.join(py_path, f"{mod_name}.py")
    with open(pyfile, "w", encoding="utf-8") as f:
        f.write(pycode)

    prev_argv = sys.argv.copy()
    prev_path = sys.path.copy()
    prev_stdout, prev_stderr = sys.stdout, sys.stderr
    prev_mypy_cache_dir = os.getenv("MYPY_CACHE_DIR")
    prev_cwd = os.getcwd()
    _Popen = subprocess.Popen
    _get_default_compiler = _base.get_default_compiler
    stream = io.StringIO()
    try:
        if quiet:
            subprocess.Popen = captured_popen(stream) # 重定向子进程的输出
        if compiler is None and _config.get("default_compiler"):
            compiler = _config["default_compiler"]
        if compiler is not None:
            def get_default_compiler(*args, **kw):
                return compiler
            _base.get_default_compiler = get_default_compiler

        if cache_dir is not None:
            os.environ["MYPY_CACHE_DIR"] = os.path.join(cache_dir,
                                                        _mypy_cache_basename)
        sys.argv[1:] = ["build_ext", "--inplace"]
        if quiet:
            sys.argv.append("-q")
        sys.stderr = sys.stdout = stream
        sys.path.append(py_path)
        os.chdir(py_path)
        try:
            setup(name = mod_name,
                ext_modules = mypycify([pyfile, "--check-untyped-defs"],
                                       opt_level=opt_level,
                                       strict_dunder_typing=strict_dunder_typing,
                                       experimental_features=experimental_features),
            )
        except SystemExit as err:
            sys.stdout, sys.stderr = prev_stdout, prev_stderr
            if err.code != 0:
                raise RuntimeError(
                    f"Nonzero exit ({err.code}), outputs:\n{stream.getvalue()}")\
                    from None

        module =  __import__(mod_name)
        module.__mypyc_aot_path__ = get_module_path(mod_name)
        return module
    finally:
        sys.argv = prev_argv
        sys.path = prev_path
        sys.stdout, sys.stderr = prev_stdout, prev_stderr
        os.chdir(prev_cwd)
        subprocess.Popen = _Popen
        _base.get_default_compiler = _get_default_compiler
        if prev_mypy_cache_dir is not None:
            os.environ["MYPY_CACHE_DIR"] = prev_mypy_cache_dir
        else: del os.environ["MYPY_CACHE_DIR"]

def mypyc_aot(pycode, prefix="mypyc_aot", cache_dir=CACHE_DIR,
              compression_method=DEFAULT_COMP_METHOD, **kw):
    try:
        if compression_method is None:
            mod = NoCompression
        else:
            mod = __import__(compression_method)
        assert mod.decompress(mod.compress(b"")) == b""
    except Exception as err:
        raise ValueError(f"""{compression_method!r} should be a module with \
reversible compress() and decompress() function:
{type(err).__name__}: {err}""") from None

    ext_suffix = sysconfig.get_config_var('EXT_SUFFIX')
    hash = base64.urlsafe_b64encode(sha256(
        f"{pycode} {ext_suffix} {kw!r} {_config!r}".encode()
        ).digest()).decode()
    cache_file = os.path.join(cache_dir, f"{hash}.{compression_method}")

    if os.path.isfile(cache_file): # 加载缓存
        with open(cache_file, "rb") as f:
            data = pickle.loads(mod.decompress(f.read()))
        py_path = os.path.join(cache_dir, "cache",
                               f"{hash}{compression_method}-mypyc")
        os.makedirs(py_path, exist_ok=True)
        filename = os.path.join(py_path, data["name"])
        with open(filename, "wb") as f:
            f.write(data["data"])
        prev_path = sys.path.copy()
        try:
            sys.path.append(py_path)
            module = __import__(data["name"].split(".")[0])
            module.__mypyc_aot_path__ = filename
            return module
        finally:
            sys.path = prev_path
    else:
        compiled = mypyc_aot_nocache(pycode, prefix, cache_dir, **kw)
        with open(compiled.__mypyc_aot_path__, "rb") as f:
            data = f.read()
        with open(cache_file, "wb") as f:
            f.write(mod.compress(
                pickle.dumps({
                    "name": os.path.basename(compiled.__mypyc_aot_path__),
                    "data": data
                })
            ))
        return compiled
